﻿#define _CRT_SECURE_NO_WARNINGS

#include <string.h>
#include <stdio.h>
#include <fcntl.h>

#include <iterator>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <stack>

std::string memory_beg;
std::string memory_end;
std::string memory_rom;
std::string memory_bin;

class CSVRow
{
public:
	std::string const & operator[](std::size_t index) const
	{
		return m_data[index];
	}
	std::size_t size() const
	{
		return m_data.size();
	}
	void read_next_row(std::istream & str)
	{
		std::string line;
		std::string cell;
		
		std::getline(str, line);

		std::stringstream lineStream(line);

		m_data.clear();
		while (std::getline(lineStream, cell, ','))
		{
			m_data.push_back(cell);
		}
		// This checks for a trailing comma with no data after it.
		if (!lineStream && cell.empty())
		{
			// If there was a trailing comma then add an empty element.
			m_data.push_back("");
		}
	}
	std::vector<std::string> const & data() const { return m_data; }
private:
	std::vector<std::string> m_data;
};

std::istream & operator>>(std::istream & str, CSVRow & data)
{
	data.read_next_row(str);
	return str;
}

int main(int argc, char * argv[])
{
	int argi = 1;

	std::string ifilename_beg;
	std::string ifilename_end;
	std::string ifilename_rom;
	std::string ifilename_csv;
	std::string ofilename_bin;

	while (argi < argc)
	{
		std::string arg = argv[argi];
		if ((arg == "--beg") && (argi + 1 < argc))
		{
			ifilename_beg = std::string(argv[++argi]);
		}
		else if ((arg == "--end") && (argi + 1 < argc))
		{
			ifilename_end = std::string(argv[++argi]);
		}
		else if ((arg == "--rom") && (argi + 1 < argc))
		{
			ifilename_rom = std::string(argv[++argi]);
		}
		else if ((arg == "--csv") && (argi + 1 < argc))
		{
			ifilename_csv = std::string(argv[++argi]);
		}
		else
		{
			ofilename_bin = std::string(arg);
		}
		++argi;
	}

	if (ifilename_beg.empty())
	{
		std::string perr = "Missing --beg filename";
		perror(perr.c_str());
	}
	if (ifilename_rom.empty())
	{
		std::string perr = "Missing --rom filename";
		perror(perr.c_str());
	}
	if (ifilename_csv.empty())
	{
		std::string perr = "Missing --csv filename";
		perror(perr.c_str());
	}
	if (ofilename_bin.empty())
	{
		std::string perr = "Missing output filename";
		perror(perr.c_str());
	}

	std::ifstream ifile_beg(ifilename_beg, std::ios::binary);
	std::ifstream ifile_rom(ifilename_rom, std::ios::binary);
	std::ifstream ifile_end(ifilename_end, std::ios::binary);
	std::ifstream ifile_csv(ifilename_csv);

	if (!ifilename_beg.empty() &&
		!ifilename_rom.empty() &&
		!ifilename_csv.empty() &&
		!ofilename_bin.empty())
	{
		if (ifile_beg) memory_beg = std::string((std::istreambuf_iterator<char>(ifile_beg)),
												 std::istreambuf_iterator<char>());
		if (memory_beg.empty())
		{
			std::string perr = "Unable to open file " + ifilename_beg;
			perror(perr.c_str());
			exit(1);
		}
		std::cout << ifilename_beg << ": " << std::to_string(memory_beg.size()) << " bytes." << std::endl;

		if (ifile_rom) memory_rom = std::string((std::istreambuf_iterator<char>(ifile_rom)),
												 std::istreambuf_iterator<char>());
		if (memory_rom.size() != 4096)
		{
			std::string perr = "Wrong size of ROM file " + ifilename_rom;
			perror(perr.c_str());
			exit(1);
		}
		std::cout << ifilename_rom << ": " << std::to_string(memory_rom.size()) << " bytes." << std::endl;

		if (!ifilename_end.empty())
		{
			if (ifile_end) memory_end = std::string((std::istreambuf_iterator<char>(ifile_end)),
													std::istreambuf_iterator<char>());
			if (memory_beg.empty())
			{
				std::string perr = "Unable to open file " + ifilename_end;
				perror(perr.c_str());
				exit(1);
			}
			std::cout << ifilename_end << ": " << std::to_string(memory_end.size()) << " bytes." << std::endl;
		}

		CSVRow csv_row;

		// Drop the header row
		ifile_csv >> csv_row;

		size_t lines = 1;

		// Read byte-to-patch rows
		while (ifile_csv >> csv_row)
		{
			size_t index = 0;
			int offset1 = 0;
			int offset2 = 0;
			int byte = 0;
			for (auto & str : csv_row.data())
			{
				bool skip_row = false;
				switch (index)
				{
				case 0: // Read offset (hex)
					offset1 = std::stoi(str, nullptr, 0);
					break;
				case 1: // Read offset (dec) and check against offset (hex)
					offset2 = std::stoi(str, nullptr, 0);
					if (offset1 != offset2)
					{
						std::string perr = "Offsets mismatch at line " + std::to_string(lines);
						perror(perr.c_str());
						exit(1);
					}
					break;
				case 2: // Read byte and patch it in binary if its offset is valid
					byte = std::stoi(str, nullptr, 0) & 255;
					if (size_t(offset1) >= memory_rom.size())
					{
						std::string perr = "Offset overflows at line " + std::to_string(lines);
						perror(perr.c_str());
						exit(1);
					}
					std::cout << ofilename_bin << ": byte " << str << " patched at offset " << std::to_string(offset1) << "." << std::endl;
					memory_rom[offset1] = char(byte);
					break;
				default: // Row has excedent data, just skip them
					skip_row = true;
					break;
				}
				if (skip_row) break;
				++index;
			}
		}

		memory_bin = memory_beg + memory_rom + memory_end;
		std::ofstream ofile_bin(ofilename_bin, std::ios::binary | std::ios::trunc);
		if (!ofile_bin)
		{
			std::string perr = "Unable to create binary " + ofilename_bin;
			perror(perr.c_str());
			exit(1);
		}
		std::cout << ofilename_bin << ": " << std::to_string(memory_bin.size()) << " bytes." << std::endl;

		ofile_bin.write(memory_bin.data(), memory_bin.size());
		ofile_bin.close();
	}
	else
	{
		printf("\n");
		printf("Usage: %s (--beg <filename>|--rom <filename>|[--end <filename>]|--csv <filename>) <filename>\n"
				"--beg <filename>: set the input filename of the begining part of the emulator to <filename>. Mandatory.\n"
				"--rom <filename>: set the input filename of the ROM file to <filename>. Mandatory.\n"
				"--end <filename>: set the input filename of the ending part of the emulator to <filename>. Optional.\n"
				"--csv <filename>: set the input filename of the CSV file to <filename>. Mandatory.\n"
				"<filename>: set the output filename of the resulting emulator binary to <filename>. Mandatory.\n", argv[0]);
	}
	return 0;
}